'''
path.py

A library designed to work with vector paths.
'''

import numpy as np
import networkx as nx

import copy
import collections

from shapely.geometry import Polygon
from scipy.spatial import cKDTree as KDTree

from ..points import plane_fit
from ..geometry import plane_transform
from ..units import _set_units
from ..util import decimal_to_digits
from ..constants import log
from ..constants import tol_path as tol


from .. import util
from .. import grouping
from .. import transformations

from . import simplify
from . import entities
from . import polygons

from .traversal import vertex_graph, closed_paths, discretize_path
from .io.export import export_path


class Path(object):
    '''
    A Path object consists of two things:
    vertices: (n,[2|3]) coordinates, stored in self.vertices
    entities: geometric primitives (lines, arcs, and circles)
              that reference indices in self.vertices
    '''

    def __init__(self,
                 entities=[],
                 vertices=[],
                 metadata=None,
                 process=True):
        '''
        entities:
            Objects which contain things like keypoints, as
            references to self.vertices
        vertices:
            (n, (2|3)) list of vertices
        '''
        self.entities = np.array(entities)
        self.vertices = vertices
        self.metadata = dict()

        if metadata.__class__.__name__ == 'dict':
            self.metadata.update(metadata)

        self._cache = util.Cache(id_function=self.md5)

        if process:
            # literally nothing will work if vertices aren't merged properly
            self.merge_vertices()

    def process(self):
        log.debug('Processing drawing')
        with self._cache:
            for func in self._process_functions():
                func()
        return self

    @property
    def vertices(self):
        return self._vertices

    @vertices.setter
    def vertices(self, values):
        self._vertices = util.tracked_array(values)

    def md5(self):
        '''
        What is an MD5 hash of the current vertex and entity arrangment.

        Not robust between loads; use self.identifier_md5 for that

        Returns
        ------------
        md5: str, MD5 of current paths
        '''
        result = self.vertices.md5()
        result += str(len(self.entities))
        return result

    @util.cache_decorator
    def paths(self):
        '''
        Sequence of closed paths, encoded by entity index.

        Returns
        ---------
        paths: (n,) sequence of (*,) int referencing self.entities
        '''
        paths = closed_paths(self.entities, self.vertices)
        return paths

    @util.cache_decorator
    def dangling(self):
        '''
        List of entities that aren't included in a closed path

        Returns
        ----------
        dangling: (n,) int, index of self.entities
        '''
        dangling = np.setdiff1d(np.arange(len(self.entities)),
                                np.hstack(self.paths))
        return dangling

    @util.cache_decorator
    def kdtree(self):
        '''
        A KDTree object holding the vertices of the path.

        Returns
        ----------
        kdtree: scipy.spatial.cKDTree object holding self.vertices
        '''

        kdtree = KDTree(self.vertices.view(np.ndarray))
        return kdtree

    @property
    def scale(self):
        '''
        What is a representitive number that reflects the magnitude
        of the world holding the paths, for numerical comparisons.

        Returns
        ----------
        scale: float, approximate size of the world holding this path
        '''
        scale = self.extents.max()
        return scale

    @util.cache_decorator
    def bounds(self):
        '''
        Return the axis aligned bounding box of the current path.

        Returns
        ----------
        bounds: (2, dimension) float, (min, max) coordinates
        '''
        points = np.array([e.bounds(self.vertices) for e in self.entities])
        points = points.reshape((-1, self.vertices.shape[1]))

        bounds = np.array([points.min(axis=0), points.max(axis=0)])

        return bounds

    @property
    def extents(self):
        return np.diff(self.bounds, axis=0)[0]

    @property
    def units(self):
        '''
        If there are units defined in self.metadata return them.

        Returns
        -----------
        units: str, current unit system
        '''
        if 'units' in self.metadata:
            return self.metadata['units']
        else:
            return None

    @units.setter
    def units(self, units):
        self.metadata['units'] = units

    def convert_units(self, desired, guess=False):
        '''
        Convert the units of the current drawing in place.

        Parameters
        -----------
        desired: str, unit system to convert to
        guess:   bool, if True will attempt to guess units
        '''
        _set_units(self, desired, guess)

    def explode(self):
        '''
        Turn every multi- segment entity into single segment entities, in- place
        '''
        new_entities = collections.deque()
        for entity in self.entities:
            new_entities.extend(entity.explode())
        self.entities = np.array(new_entities)

    def fill_gaps(self, max_distance=np.inf):
        '''
        Find vertexes with degree 1 and try to connect them to other
        vertices of degree 1, in place.

        Parameters
        ----------
        max_distance: float, connect vertices up to this distance.
                      Default is infinity, but something like path.scale/100
                      may make more sense.
        '''

        broken = np.array(
            [k for k, v in dict(self.vertex_graph.degree()).items() if v == 1])
        if len(broken) < 2:
            return

        distance, node = KDTree(self.vertices[broken]).query(
            self.vertices[broken], k=2)
        edges = broken[node]
        ok = np.logical_and(distance[:, 1] < max_distance, [
                            not self.vertex_graph.has_edge(*i) for i in edges])

        self.entities = np.append(self.entities,
                                  [entities.Line(i) for i in edges[ok]])

    @property
    def is_closed(self):
        '''
        Are all entities connected to other entities.

        Returns
        -----------
        closed: every entity is connected at its ends
        '''
        closed = all(i == 2 for i in dict(self.vertex_graph.degree()).values())
        return closed

    @util.cache_decorator
    def vertex_graph(self):
        '''
        Return a networkx.Graph object for the entity connectiviy

        graph: networkx.Graph object, holding vertex indexes
        '''
        graph, closed = vertex_graph(self.entities)
        return graph

    def apply_transform(self, transform):
        '''
        Apply a transformation matrix to the current path in- place

        Parameters
        -----------
        transform: (dimension + 1, dimension + 1) float, homogenous
                   transformation matrix
        '''
        self.vertices = transformations.transform_points(self.vertices,
                                                         transform)
    def apply_scale(self, scale):
        '''
        Apply a transformation matrix to the current path in- place

        Parameters
        -----------
        transform: (dimension + 1, dimension + 1) float, homogenous
                   transformation matrix
        '''
        dimension = self.vertices.shape[1]
        matrix = np.eye(dimension+1)
        matrix[:dimension,:dimension] *= float(scale)
        self.apply_transform(matrix)

    def rezero(self):
        '''
        Translate so that every vertex is positive, in-place.
        '''
        self.vertices -= self.vertices.min(axis=0)

    def merge_vertices(self):
        '''
        Merges vertices which are identical and replaces references in place.
        '''
        digits = decimal_to_digits(tol.merge * self.scale, min_digits=1)
        unique, inverse = grouping.unique_rows(self.vertices, digits=digits)
        self.vertices = self.vertices[unique]
        for entity in self.entities:
            # if we merged duplicate vertices, the entity may contain
            # multiple references to the same vertex
            entity.points = grouping.merge_runs(inverse[entity.points])

    def replace_vertex_references(self, replacement_dict):
        for entity in self.entities:
            entity.rereference(replacement_dict)

    def remove_entities(self, entity_ids):
        '''
        Remove entities by their index.
        '''
        if len(entity_ids) == 0:
            return
        kept = np.setdiff1d(np.arange(len(self.entities)), entity_ids)
        self.entities = np.array(self.entities)[kept]

    def remove_invalid(self):
        valid = np.array([i.is_valid for i in self.entities], dtype=np.bool)
        self.entities = self.entities[valid]

    def remove_duplicate_entities(self):
        entity_hashes = np.array([i.hash for i in self.entities])
        unique, inverse = grouping.unique_rows(entity_hashes)
        if len(unique) != len(self.entities):
            self.entities = np.array(self.entities)[unique]

    def referenced_vertices(self):
        referenced = collections.deque()
        for entity in self.entities:
            referenced.extend(entity.points)
        return np.array(referenced)

    def remove_unreferenced_vertices(self):
        '''
        Removes all vertices which aren't used by an entity
        Reindexes vertices from zero, and replaces references
        '''
        referenced = self.referenced_vertices()
        unique_ref = np.int_(np.unique(referenced))
        replacement_dict = dict()
        replacement_dict.update(np.column_stack((unique_ref,
                                                 np.arange(len(unique_ref)))))
        self.replace_vertex_references(replacement_dict)
        self.vertices = self.vertices[[unique_ref]]

    def discretize_path(self, path):
        '''
        Return a (n, dimension) list of vertices.
        Samples arcs/curves to be line segments
        '''
        discrete = discretize_path(self.entities,
                                   self.vertices,
                                   path,
                                   scale=self.scale)
        return discrete

    def paths_to_splines(self, path_indexes=None, smooth=.0002):
        '''
        Convert paths into b-splines.

        Parameters
        -----------
        path_indexes: (n) int list of indexes for self.paths
        smooth:       float, how much the spline should smooth the curve
        '''
        if path_indexes is None:
            path_indexes = np.arange(len(self.paths))
        entities_keep = np.ones(len(self.entities), dtype=np.bool)
        new_vertices = collections.deque()
        new_entities = collections.deque()
        for i in path_indexes:
            path = self.paths[i]
            discrete = self.discrete[i]
            entity, vertices = simplify.points_to_spline_entity(discrete)
            entity.points += len(self.vertices) + len(new_vertices)
            new_vertices.extend(vertices)
            new_entities.append(entity)
            entities_keep[path] = False
        self.entities = np.append(self.entities[entities_keep],
                                  new_entities)
        self.vertices = np.vstack((self.vertices,
                                   np.array(new_vertices)))

    def export(self, file_obj=None, file_type='dict'):
        return export_path(self,
                           file_type=file_type,
                           file_obj=file_obj)

    def to_dict(self):
        export_dict = self.export(file_type='dict')
        return export_dict

    def copy(self):
        return copy.deepcopy(self)

    def show(self):
        if self.is_closed:
            self.plot_discrete(show=True)
        else:
            self.plot_entities(show=True)

    def __add__(self, other):
        new_entities = copy.deepcopy(other.entities)
        for entity in new_entities:
            entity.points += len(self.vertices)
        new_entities = np.append(copy.deepcopy(self.entities), new_entities)

        new_vertices = np.vstack((self.vertices, other.vertices))
        new_meta = copy.deepcopy(self.metadata)
        new_meta.update(other.metadata)

        new_path = self.__class__(entities=new_entities,
                                  vertices=new_vertices,
                                  metadata=new_meta)
        return new_path


class Path3D(Path):

    def _process_functions(self):
        return [self.merge_vertices,
                self.remove_duplicate_entities,
                self.remove_unreferenced_vertices,
                self.generate_closed_paths,
                self.generate_discrete]

    @util.cache_decorator
    def discrete(self):
        discrete = list(map(self.discretize_path, self.paths))
        return discrete

    def to_planar(self, to_2D=None, normal=None, check=True):
        '''
        Check to see if current vectors are all coplanar.

        If they are, return a Path2D and a transform which will
        transform the 2D representation back into 3 dimensions

        Parameters
        -----------
        to_2D: (4,4) float, transformation matrix to apply.
                     If not passed a plane will be fitted to vertices.
        normal: (3,) float, normal of plane which is only used if to_2D is not specified
        check:  bool, raise a ValueError if the points aren't coplanar after
                      being transformed

        Returns
        -----------
        planar: Path2D object, current path transformed onto a plane
        to_3D:  (4,4), transformation matrix to move planar back into 3D space
        '''
        if to_2D is None:
            C, N = plane_fit(self.vertices)
            if normal is not None:
                N *= np.sign(np.dot(N, normal))
                N = normal
            to_2D = plane_transform(C, N)

        flat = transformations.transform_points(self.vertices, to_2D)

        if check and np.any(np.std(flat[:, 2]) > tol.planar):
            log.error('points have z with deviation %f', np.std(flat[:, 2]))
            raise NameError('Points aren\'t planar!')

        planar = Path2D(entities=copy.deepcopy(self.entities),
                        vertices=flat[:, 0:2],
                        metadata=copy.deepcopy(self.metadata))
        to_3D = np.linalg.inv(to_2D)

        return planar, to_3D

    def scene(self):
        '''
        Get a scene object containing the current Path3D object.

        Returns
        --------
        scene: trimesh.scene.Scene object containing current path
        '''
        from ..scene import Scene
        scene = Scene(self)
        return scene

    def show(self):
        '''
        Show the current Path3D object.
        '''
        self.scene().show()

    def plot_discrete(self, show=False):
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D
        fig = plt.figure()
        axis = fig.add_subplot(111, projection='3d')
        for discrete in self.discrete:
            axis.plot(*discrete.T)
        if show:
            plt.show()

    def plot_entities(self, show=False):
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D
        fig = plt.figure()
        axis = fig.add_subplot(111, projection='3d')
        for entity in self.entities:
            vertices = entity.discrete(self.vertices)
            axis.plot(*vertices.T)
        if show:
            plt.show()


class Path2D(Path):

    def _process_functions(self):
        return [self.merge_vertices,
                self.remove_duplicate_entities,
                self.remove_unreferenced_vertices]

    def apply_obb(self):
        if len(self.root) == 1:
            matrix, bounds = polygons.polygon_obb(
                self.polygons_closed[self.root[0]])
            self.apply_transform(matrix)
            return matrix
        else:
            raise ValueError('Not implemented for multibody geometry')

    @property
    def body_count(self):
        return len(self.root)

    def to_3D(self):
        '''
        Convert 2D path to 3D path on the XY plane.

        Returns
        -----------
        path_3D: Path3D version of current path
        '''
        vertices_new = np.column_stack((copy.deepcopy(self.vertices),
                                        np.zeros(len(self.vertices))))
        path_3D = Path3D(entities=copy.deepcopy(self.entities),
                         vertices=vertices_new,
                         metadata=copy.deepcopy(self.metadata))
        return path_3D

    @util.cache_decorator
    def polygons_full(self):
        '''
        A list of shapely.geometry.Polygon objects with corresponding interiors
        pulled by looking at which closed polygons enclose which other polygons.

        Returns
        ---------
        full: list of shapely.geometry.Polygon objects
        '''
        full = [None] * len(self.enclosure_shell)

        for index, (shell, hole) in enumerate(self.enclosure_shell.items()):
            if not self.path_valid[shell]:
                continue
            # mask indexes of holes by valid polygons
            # this will silently discard invalid holes
            hole_idx = hole[self.path_valid[hole]]
            # closed curves which represent holes
            hole_poly = self.polygons_closed[hole_idx]
            # generate a new polygon with shell and holes
            polygon = Polygon(
                shell=self.polygons_closed[shell].exterior.coords, holes=[
                    p.exterior.coords for p in hole_poly])
            full[index] = polygon
        return full

    @util.cache_decorator
    def area(self):
        '''
        Return the area of the polygons interior.
        '''
        area = np.sum([i.area for i in self.polygons_full])
        return area

    @util.cache_decorator
    def length(self):
        '''
        The total discretized length of every entity.

        Returns
        --------
        length: float, summed length of every entity
        '''
        length = sum(i.length(self.vertices) for i in self.entities)
        return length

    def extrude(self, height, **kwargs):
        '''
        Extrude the current 2D path into a 3D mesh.

        Parameters
        ----------
        height: float, how far to extrude the profile
        kwargs: passed directly to meshpy.triangle.build:
                triangle.build(mesh_info,
                               verbose=False,
                               refinement_func=None,
                               attributes=False,
                               volume_constraints=True,
                               max_volume=None,
                               allow_boundary_steiner=True,
                               allow_volume_steiner=True,
                               quality_meshing=True,
                               generate_edges=None,
                               generate_faces=False,
                               min_angle=None)
        Returns
        --------
        mesh: trimesh object representing extruded polygon
        '''
        from ..primitives import Extrusion
        result = [Extrusion(polygon=i, height=height, **kwargs)
                  for i in self.polygons_full]
        if len(result) == 1:
            return result[0]
        return result

    def medial_axis(self, resolution=None, clip=None):
        '''
        Find the approximate medial axis based
        on a voronoi diagram of evenly spaced points on the boundary of the polygon.

        Parameters
        ----------
        resolution: target distance between each sample on the polygon boundary
        clip:       [minimum number of samples, maximum number of samples]
                    specifying a very fine resolution can cause the sample count to
                    explode, so clip specifies a minimum and maximum number of samples
                    to use per boundary region. To not clip, this can be specified as:
                    [0, np.inf]

        Returns
        ----------
        medial:     Path2D object
        '''
        if 'medial' in self._cache:
            return self._cache.get('medial')

        if resolution is None:
            resolution = self.scale / 1000.0

        medials = [polygons.medial_axis(i, resolution, clip)
                   for i in self.polygons_full]
        medials = np.sum(medials)
        return self._cache.set(key='medial',
                               value=medials)

    def connected_paths(self, path_id, include_self=False):
        '''
        Given an index of self.paths, find other paths which overlap with
        that path.

        Parameters
        -----------
        path_id:      int, index of self.paths
        include_self: bool, should the result include path_id or not

        Returns
        -----------
        path_ids: (n,) int, indexes of self.paths that overlap input path_id
        '''
        if len(self.root) == 1:
            path_ids = np.arange(len(self.polygons_closed))
        else:
            path_ids = list(nx.node_connected_component(self.enclosure,
                                                        path_id))
        if include_self:
            return np.array(path_ids)
        return np.setdiff1d(path_ids, [path_id])

    def simplify(self):
        '''
        Return a version of the current path with colinear segments
        merged, and circles entities replacing segmented circular paths.

        Returns
        ---------
        simplified: Path2D object
        '''
        return simplify.simplify_basic(self)

    def split(self):
        '''
        If the current Path2D consists of n 'root' curves,
        split them into a list of n Path2D objects

        Returns
        ----------
        split: (n,) list of Path2D objects
        '''
        if self.root is None or len(self.root) == 0:
            split = []
        elif len(self.root) == 1:
            split = [copy.deepcopy(self)]
        else:
            split = [None] * len(self.root)
            for i, root in enumerate(self.root):
                connected = self.connected_paths(root, include_self=True)
                new_root = np.nonzero(connected == root)[0]
                new_entities = collections.deque()
                new_paths = collections.deque()
                new_metadata = copy.deepcopy(self.metadata)
                new_metadata['split_2D'] = i

                for path in self.paths[connected]:
                    new_paths.append(np.arange(len(path)) + len(new_entities))
                    new_entities.extend(path)
                new_entities = np.array(new_entities)
                # prevents the copying from nuking our cache
                with self._cache:
                    split[i] = Path2D(entities=copy.deepcopy(self.entities[new_entities]),
                                      vertices=copy.deepcopy(self.vertices),
                                      metadata=new_metadata)
                    split[i]._cache.update(
                        {
                            'paths': np.array(new_paths),
                            'polygons_closed': self.polygons_closed[connected],
                            'path_valid': self.path_valid[connected],
                            'discrete': self.discrete[connected],
                            'root': new_root})
        [i._cache.id_set() for i in split]
        self._cache.id_set()
        return np.array(split)

    def plot_discrete(self, show=False, transform=None, axes=None):
        '''
        Plot the closed curves of the path.
        '''
        import matplotlib.pyplot as plt
        plt.axes().set_aspect('equal', 'datalim')

        def plot_transformed(vertices, color='g'):
            if transform is None:
                if axes is None:
                    plt.plot(*vertices.T, color=color)
                else:
                    axes.plot(*vertices.T, color=color)
            else:
                transformed = transformations.transform_points(
                    vertices, transform)
                plt.plot(*transformed.T, color=color)
        for i, polygon in enumerate(self.polygons_closed):
            color = ['g', 'k'][i in self.root]
            plot_transformed(np.array(polygon.boundary.coords), color=color)
        if show:
            plt.show()

    def plot_entities(self, show=False):
        '''
        Plot the entities of the path, with no notion of topology
        '''
        import matplotlib.pyplot as plt
        plt.axes().set_aspect('equal', 'datalim')
        eformat = {'Line0': {'color': 'g', 'linewidth': 1},
                   'Line1': {'color': 'y', 'linewidth': 1},
                   'Arc0': {'color': 'r', 'linewidth': 1},
                   'Arc1': {'color': 'b', 'linewidth': 1},
                   'Bezier0': {'color': 'k', 'linewidth': 1},
                   'BSpline0': {'color': 'm', 'linewidth': 1},
                   'BSpline1': {'color': 'm', 'linewidth': 1}}
        for entity in self.entities:
            discrete = entity.discrete(self.vertices)
            e_key = entity.__class__.__name__ + str(int(entity.closed))
            plt.plot(discrete[:, 0],
                     discrete[:, 1],
                     **eformat[e_key])
        if show:
            plt.show()

    @property
    def identifier(self):
        '''
        A unique identifier for the path.

        Returns
        ---------
        identifier: (5,) float, unique identifier
        '''
        if len(self.polygons_full) != 1:
            raise TypeError('Identifier only valid for single body')
        return polygons.polygon_hash(self.polygons_full[0])

    @util.cache_decorator
    def identifier_md5(self):
        '''
        Return an MD5 of the identifier
        '''
        return util.md5_array(self.identifier, digits=4)

    @property
    def path_valid(self):
        '''
        Returns
        ----------
        path_valid: (n,) bool, indexes of self.paths self.polygons_closed
                         which are valid polygons
        '''
        exists = self.polygons_closed
        return self._cache.get('path_valid')

    @property
    def discrete(self):
        '''
        Returns
        ---------
        discrete: sequence of (n,2) float, correspond to self.paths
        '''
        if 'discrete' not in self._cache:
            test = self.polygons_closed
        return self._cache['discrete']

    @util.cache_decorator
    def polygons_closed(self):
        '''
        Cycles in the vertex graph, as shapely.geometry.Polygons.
        These are polygon objects for every closed circuit, with no notion
        of whether a polygon is a hole or an area. Every polygon in this
        list will have an exterior, but NO interiors.

        Returns
        ---------
        polygons_closed: (n,) list of shapely.geometry.Polygon objects
        '''
        def reverse_path(path):
            for entity in self.entities[path]:
                entity.reverse()
            return path[::-1]

        with self._cache:
            discretized = [None] * len(self.paths)
            new_polygon = [None] * len(self.paths)
            valid = np.zeros(len(self.paths), dtype=np.bool)
            for i, path in enumerate(self.paths):
                discrete = discretize_path(self.entities,
                                           self.vertices,
                                           path,
                                           scale=self.scale)
                candidate = polygons.path_to_polygon(discrete,
                                                     scale=self.scale)
                if candidate is None:
                    continue
                if type(candidate).__name__ == 'MultiPolygon':
                    area_ok = np.array([i.area for i in candidate]) > tol.zero
                    if area_ok.sum() == 1:
                        candidate = candidate[np.nonzero(area_ok)[0][0]]
                    else:
                        continue
                if not candidate.exterior.is_ccw:
                    log.debug('Clockwise polygon detected, correcting!')
                    self.paths[i] = reverse_path(path)
                    candidate = Polygon(
                        np.array(candidate.exterior.coords)[::-1])
                new_polygon[i] = candidate
                valid[i] = True
                discretized[i] = discrete

            new_polygon = np.array(new_polygon)[valid]
            discretized = np.array(discretized)

        self._cache.set('discrete', discretized)
        self._cache.set('path_valid', valid)

        return new_polygon

    @util.cache_decorator
    def root(self):
        '''
        Which indexes of self.paths/self.polygons_closed are root curves.
        Also known as 'shell' or 'exterior.

        Returns
        ---------
        root: (n,) int, list of indexes
        '''
        populate = self.enclosure_directed
        return self._cache['root']

    @util.cache_decorator
    def enclosure(self):
        '''
        Networkx Graph object of polygon enclosure.
        '''
        with self._cache:
            undirected = self.enclosure_directed.to_undirected()
        return undirected

    @util.cache_decorator
    def enclosure_directed(self):
        '''
        Networkx DiGraph of polygon enclosure
        '''
        root, enclosure = polygons.polygons_enclosure_tree(
            self.polygons_closed)
        self._cache.set('root', root)
        return enclosure

    @util.cache_decorator
    def enclosure_shell(self):
        '''
        A dictionary of path indexes which are 'shell' paths, and values
        of 'hole' paths.

        Returns
        ----------
        corresponding: dict, {index of self.paths of shell : [indexes of holes]}
        '''
        pairs = [(r, self.connected_paths(r, include_self=False))
                 for r in self.root]
        # OrderedDict to maintain corresponding order
        corresponding = collections.OrderedDict(pairs)
        return corresponding



'''
This module provides GLTF 2.0 exports 
'''

import json

import numpy as np


def export_gltf(scene):
    '''
    Export a scene object as a GLTF directory.
    
    This has the advantage of putting each mesh into a separate file (buffer)
    as opposed to one large file, but means multiple files need to be tracked.

    Parameters
    -----------
    scene: trimesh.Scene object

    Returns
    ----------
    export: dict, {file name : file data}
    '''

    tree, buffer_items = _gltf_structure(scene)

    buffers = []
    views = []
    files = {}
    for i, name in zip(range(0, len(buffer_items), 2),
                       scene.geometry.keys()):

        # create the buffer views
        current_pos = 0
        for j in range(2):
            current_item = buffer_items[i + j]
            views.append({"buffer": len(buffers),
                          "byteOffset": current_pos,
                          "byteLength": len(current_item)})
            current_pos += len(current_item)

        # the data is just appended
        buffer_data = bytes().join(buffer_items[i:i + 2])
        buffer_name = 'mesh_' + name + '.bin'
        buffers.append({'uri': buffer_name,
                        'byteLength': len(buffer_data)})
        files[buffer_name] = buffer_data

    tree['buffers'] = buffers
    tree['bufferViews'] = views

    files['model.gltf'] = json.dumps(tree).encode('utf-8')
    return files


def export_glb(scene):
    '''
    Export a scene as a binary GLTF (GLB) file.

    Parameters
    ------------
    scene: trimesh.Scene object

    Returns
    ----------
    exported: bytes, exported result
    '''

    tree, buffer_items = _gltf_structure(scene)

    # A bufferView is a slice of a file
    views = []
    # create the buffer views
    current_pos = 0
    for current_item in buffer_items:
        views.append({"buffer": 0,
                      "byteOffset": current_pos,
                      "byteLength": len(current_item)})
        current_pos += len(current_item)

        # the data is just appended
    buffer_data = bytes().join(buffer_items)

    tree['buffers'] = [{'byteLength': len(buffer_data)}]
    tree['bufferViews'] = views

    # export the tree to JSON for the content of the file
    content = json.dumps(tree)
    # add spaces to content, so the start of the data
    # is 4 byte aligned as per spec
    content += ((len(content) + 20) % 4) * ' '
    content = content.encode('utf-8')

    # the initial header of the file
    header = np.array([1179937895,  # magic, turns into glTF
                       2,  # GLTF version
                       # length is the total length of the Binary glTF
                       # including Header and all Chunks, in bytes.
                       len(content) + len(buffer_data) + 28,
                       # contentLength is the length, in bytes,
                       # of the glTF content (JSON)
                       len(content),
                       # magic number which is 'JSON'
                       1313821514],
                      dtype=np.uint32)

    # the header of the binary data section
    bin_header = np.array([len(buffer_data),
                           0x004E4942],
                          dtype=np.uint32)

    exported = (header.tostring() +
                content +
                bin_header.tostring() +
                buffer_data)

    return exported


def _gltf_structure(scene):
    # we are defining a single scene, and will be setting the
    # world node to the 0th index
    tree = {'scene': 0,
            'scenes': [{'nodes': [0]}],
            "asset": {"version": "2.0",
                      "generator": "github.com/mikedh/trimesh"},
            'accessors': [],
            'meshes': []}

    # GLTF references meshes by index, so store them here
    mesh_index = {name: i for i, name in enumerate(scene.geometry.keys())}
    # grab the flattened scene graph in GLTF's format
    nodes = scene.graph.to_gltf(mesh_index=mesh_index)
    tree.update(nodes)

    buffer_items = []
    for name, mesh in scene.geometry.items():
        # meshes reference accessor indexes
        tree['meshes'].append({"name": name,
                               "primitives": [
                                   {"attributes": {"POSITION": len(tree['accessors']) + 1},
                                    "indices": len(tree['accessors']),
                                    "mode": 4}]})  # mode 4 is GL_TRIANGLES

        # accessors refer to data locations
        # mesh faces are stored as flat list of integers
        tree['accessors'].append({"bufferView": len(buffer_items),
                                  "componentType": 5125,
                                  "count": len(mesh.faces) * 3,
                                  "max": [int(mesh.faces.max())],
                                  "min": [0],
                                  "type": "SCALAR"})

        # the vertex accessor
        tree['accessors'].append({"bufferView": len(buffer_items) + 1,
                                  "componentType": 5126,
                                  "count": len(mesh.vertices),
                                  "type": "VEC3",
                                  "max": mesh.vertices.max(axis=0).tolist(),
                                  "min": mesh.vertices.min(axis=0).tolist()})

        # convert to the correct dtypes
        # 5126 is a float32
        # 5125 is an unsigned 32 bit integer
        # add faces, then vertices
        buffer_items.append(mesh.faces.astype(np.uint32).tostring())
        buffer_items.append(mesh.vertices.astype(np.float32).tostring())

    return tree, buffer_items



import time
import collections

import numpy as np
import networkx as nx

from .. import util
from .. import transformations


class TransformForest:

    def __init__(self, base_frame='world'):
        self.transforms = EnforcedForest()
        self.base_frame = base_frame

        self._paths = {}
        self._updated = time.time()

        self._cache = util.Cache(id_function=self.md5)

    def update(self,
               frame_to,
               frame_from=None,
               **kwargs):
        '''
        Update a transform in the tree.

        Parameters
        ---------
        frame_from: hashable object, usually a string (eg 'world').
                    If left as None it will be set to self.base_frame
        frame_to:   hashable object, usually a string (eg 'mesh_0')

        Additional kwargs (can be used in combinations)
        ---------
        matrix:      (4,4) array
        quaternion:  (4) quaternion
        axis:        (3) array
        angle:       float, radians
        translation: (3) array

        geometry: Geometry object name
        '''
        if frame_from is None:
            frame_from = self.base_frame
        matrix = kwargs_to_matrix(**kwargs)

        attr = {'matrix': matrix,
                'time': time.time()}

        if 'geometry' in kwargs:
            attr['geometry'] = kwargs['geometry']

        changed = self.transforms.add_edge(frame_from,
                                           frame_to,
                                           **attr)
        if 'geometry' in kwargs:
            nx.set_node_attributes(self.transforms,
                                   name='geometry',
                                   values={frame_to: kwargs['geometry']})
        if changed:
            self._paths = {}
        self._updated = time.time()

    def md5(self):
        '''
        MD5 of transforms.

        Currently only hashing update time.
        '''
        result = str(int(self._updated * 1000)) + str(self.base_frame)
        return result

    def to_flattened(self, base_frame=None):
        '''
        Export the current transform graph as a flattened
        '''
        if base_frame is None:
            base_frame = self.base_frame

        flat = {}
        for node in self.nodes:
            if node == base_frame:
                continue
            transform, geometry = self.get(frame_to=node,
                                           frame_from=base_frame)
            flat[node] = {'transform': transform.tolist(),
                          'geometry': geometry}
        return flat

    def to_gltf(self, mesh_index):
        '''
        Export a list of transforms as the 'nodes' section of a GLTF dict.

        Will flatten tree.

        Returns
        --------
        gltf: dict, with keys:
                  'nodes': list of dicts
        '''
        gltf = collections.deque()

        for node in self.nodes:
            if node == self.base_frame:
                continue
            transform, geometry = self.get(frame_to=node,
                                           frame_from=self.base_frame)
            if geometry is None:
                continue

            gltf.append({'matrix': transform.T.reshape(-1).tolist(),
                         'mesh': mesh_index[geometry],
                         'name': node})

        # we have flattened tree, so all nodes will be child of world
        gltf.appendleft({'name': self.base_frame,
                         'children': list(range(1, 1 + len(gltf)))})
        result = {'nodes': list(gltf)}
        return result

    def to_edgelist(self):
        '''
        Export the current transforms as a list of edge tuples, with
        each tuple having the format:
        (node_a, node_b, {metadata})

        Returns
        -------
        edgelist: (n,) list of tuples
        '''
        # wrapped in a list for nx 2.0
        export = list(nx.to_edgelist(self.transforms))
        for e in export:
            e[2]['matrix'] = np.array(e[2]['matrix']).tolist()
        return export

    def load(self, edgelist):
        for edge in edgelist:
            self.transforms.add_edge(edge[0], edge[1], **edge[2])

    @util.cache_decorator
    def nodes(self):
        return np.array(self.transforms.nodes())

    @util.cache_decorator
    def nodes_geometry(self):
        nodes = [i for i in self.nodes if 'geometry' in self.transforms.node[i]]
        return np.array(nodes)

    def get(self,
            frame_to,
            frame_from=None):
        '''
        Get the transform from one frame to another, assuming they are connected
        in the transform tree.

        If the frames are not connected a NetworkXNoPath error will be raised.

        Parameters
        ---------
        frame_from: hashable object, usually a string (eg 'world').
                    If left as None it will be set to self.base_frame
        frame_to:   hashable object, usually a string (eg 'mesh_0')

        Returns
        ---------
        transform:  (4,4) homogenous transformation matrix
        '''

        if frame_from is None:
            frame_from = self.base_frame

        cache_key = str(frame_from) + ':' + str(frame_to)
        cached = self._cache[cache_key]
        if cached is not None:
            return cached

        transform = np.eye(4)
        path = self._get_path(frame_from, frame_to)

        for i in range(len(path) - 1):
            data, direction = self.transforms.get_edge_data_direction(
                path[i], path[i + 1])
            matrix = data['matrix']
            if direction < 0:
                matrix = np.linalg.inv(matrix)
            transform = np.dot(transform, matrix)

        geometry = None
        if 'geometry' in self.transforms.node[frame_to]:
            geometry = self.transforms.node[frame_to]['geometry']

        self._cache[cache_key] = (transform, geometry)

        return transform, geometry

    def show(self):
        '''
        Plot the graph layout of the scene.
        '''
        import matplotlib.pyplot as plt
        nx.draw(self.transforms, with_labels=True)
        plt.show()

    def to_svg(self):
        '''
        '''
        from ..graph import graph_to_svg
        return graph_to_svg(self.transforms)

    def __contains__(self, key):
        return key in self.transforms.node

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        value = np.asanyarray(value)
        if value.shape != (4, 4):
            raise ValueError('Matrix must be specified!')
        return self.update(key, matrix=value)

    def clear(self):
        self.transforms = EnforcedForest()
        self._paths = {}
        self._updated = time.time()

    def _get_path(self,
                  frame_from,
                  frame_to):
        '''
        Find a path between two frames, either from cached paths or
        from the transform graph.

        Parameters
        ---------
        frame_from: a frame key, usually a string
                    eg, 'world'
        frame_to:   a frame key, usually a string
                    eg, 'mesh_0'

        Returns
        ----------
        path: (n) list of frame keys
              eg, ['mesh_finger', 'mesh_hand', 'world']
        '''
        key = (frame_from, frame_to)
        if not (key in self._paths):
            path = self.transforms.shortest_path_undirected(frame_from,
                                                            frame_to)
            self._paths[key] = path
        return self._paths[key]


class EnforcedForest(nx.DiGraph):
    '''
    A subclass of networkx.DiGraph that will raise an error if an
    edge is added which would make the DiGraph not a forest or tree.
    '''

    def __init__(self, *args, **kwargs):
        self.flags = {'strict': False,
                      'assert_forest': False}

        for k, v in self.flags.items():
            if k in kwargs:
                self.flags[k] = bool(kwargs[k])
                kwargs.pop(k, None)

        super(self.__class__, self).__init__(*args, **kwargs)
        # keep a second parallel but undirected copy of the graph
        # all of the networkx methods for turning a directed graph
        # into an undirected graph are quite slow, so we do minor bookkeeping
        self._undirected = nx.Graph()

    def add_edge(self, u, v, *args, **kwargs):
        changed = False
        if u == v:
            if self.flags['strict']:
                raise ValueError('Edge must be between two unique nodes!')
            return changed
        if self._undirected.has_edge(u, v):
            self.remove_edges_from([[u, v], [v, u]])
        elif len(self.nodes()) > 0:
            try:
                path = nx.shortest_path(self._undirected, u, v)
                if self.flags['strict']:
                    raise ValueError(
                        'Multiple edge path exists between nodes!')
                self.disconnect_path(path)
                changed = True
            except (nx.NetworkXError, nx.NetworkXNoPath, nx.NetworkXException):
                pass
        self._undirected.add_edge(u, v)
        super(self.__class__, self).add_edge(u, v, *args, **kwargs)

        if self.flags['assert_forest']:
            # this is quite slow but makes very sure structure is correct
            # so is mainly used for testing
            assert nx.is_forest(nx.Graph(self))

        return changed

    def add_edges_from(self, *args, **kwargs):
        raise ValueError('EnforcedTree requires add_edge method to be used!')

    def add_path(self, *args, **kwargs):
        raise ValueError('EnforcedTree requires add_edge method to be used!')

    def remove_edge(self, *args, **kwargs):
        super(self.__class__, self).remove_edge(*args, **kwargs)
        self._undirected.remove_edge(*args, **kwargs)

    def remove_edges_from(self, *args, **kwargs):
        super(self.__class__, self).remove_edges_from(*args, **kwargs)
        self._undirected.remove_edges_from(*args, **kwargs)

    def disconnect_path(self, path):
        ebunch = np.array([[path[0], path[1]]])
        ebunch = np.vstack((ebunch, np.fliplr(ebunch)))
        self.remove_edges_from(ebunch)

    def shortest_path_undirected(self, u, v):
        path = nx.shortest_path(self._undirected, u, v)
        return path

    def get_edge_data_direction(self, u, v):
        if self.has_edge(u, v):
            direction = 1
        elif self.has_edge(v, u):
            direction = -1
        else:
            raise ValueError('Edge doesnt exist!')
        data = self.get_edge_data(*[u, v][::direction])
        return data, direction


def path_to_edges(path):
    '''
    Turn an (n) path into a (2(n-1)) set of edges
    '''
    return np.column_stack((path, path)).reshape(-1)[1:-1].reshape((-1, 2))


def kwargs_to_matrix(**kwargs):
    '''
    Turn a set of keyword arguments into a transformation matrix.
    '''
    matrix = np.eye(4)
    if 'matrix' in kwargs:
        # a matrix takes precedence over other options
        matrix = kwargs['matrix']
    elif 'quaternion' in kwargs:
        matrix = transformations.quaternion_matrix(kwargs['quaternion'])
    elif ('axis' in kwargs) and ('angle' in kwargs):
        matrix = transformations.rotation_matrix(kwargs['angle'],
                                                 kwargs['axis'])
    else:
        raise ValueError('Couldn\'t update transform!')

    if 'translation' in kwargs:
        # translation can be used in conjunction with any of the methods of
        # specifying transforms. In the case a matrix and translation are passed,
        # we add the translations together rather than picking one.
        matrix[0:3, 3] += kwargs['translation']
    return matrix



import numpy as np

import collections
import json
import re

from .. import util
from .. import geometry


def load_off(file_obj, file_type=None):
    '''
    Load an OFF file into the kwargs for a Trimesh constructor


    Parameters
    ----------
    file_obj: file object containing an OFF file
    file_type: not used

    Returns
    ----------
    loaded: dict with kwargs for Trimesh constructor (vertices, faces)

    '''
    header_string = file_obj.readline()
    if hasattr(header_string, 'decode'):
        header_string = header_string.decode('utf-8')
    header_string = header_string.strip()

    if not header_string == 'OFF':
        raise NameError('Not an OFF file! Header was ' + header_string)

    header = np.array(file_obj.readline().split()).astype(int)
    blob = np.array(file_obj.read().split())
    data_ok = np.sum(header * [3, 4, 0]) == len(blob)
    if not data_ok:
        raise NameError('Incorrect number of vertices or faces!')

    vertices = blob[0:(header[0] * 3)].astype(float).reshape((-1, 3))
    faces = blob[(header[0] * 3):].astype(int).reshape((-1, 4))[:, 1:]

    return {'vertices': vertices,
            'faces': faces}


def load_wavefront(file_obj, file_type=None):
    '''
    Loads an ascii Wavefront OBJ file_obj into kwargs
    for the Trimesh constructor.

    Discards texture normals and vertex color information.


    Parameters
    ----------
    file_obj: file object containing a wavefront file
    file_type: not used

    Returns
    ----------
    loaded: dict with kwargs for Trimesh constructor (vertices, faces)
    '''
    text_original = file_obj.read()
    if hasattr(text_original, 'decode'):
        text_original = text_original.decode('utf-8')
    # get rid of stupid newlines
    text_original = text_original.replace(
        '\r\n', '\n').replace(
        '\r', '\n') + ' \n'

    # for faces, remove the '/' notation in the raw text
    # the regex does the following:
    # find charecter '/' and then any non- whitespace charecter,
    # up to a newline or space, then stop. Example:
    # test = "f 233/233/233 12//12//12\nf 233/233/233 12//12//12 "
    # In [0]: re.split('/\S*[ \n]', test)
    # Out[0]: ['f 233', '12', 'f 233', '12', '']
    # we then re-join it into a larger string so we
    # can split by just whitespace
    # we add a space before every newline to make things easy on ourselves
    text = ' '.join(re.split('/\S* ',
                             text_original.replace('\n', ' \n')))

    # remove all comments
    # regex:
    # # : comments start with pounds
    # .*: zero or more of any charecter
    # \n: up to a newline
    text = re.sub('#.*\n', '\n', text)

    # more impenetrable regexes
    # this one to pulls faces from directly from the text
    # - find the 'f' char
    # - followed by one or more spaces
    # - followed by one or more 0-9 digits
    # - followed by one or more spaces
    # - (repeat for exactly 3 integers for tris, 4 for quads)
    # - followed by zero or more spaces
    # - followed by exactly one newline
    re_tris = 'f +\d+ +\d+ +\d+ *\n'
    re_quad = 'f +\d+ +\d+ +\d+ +\d+ *\n'

    # Split the file on object lines -- any line that begins with an 'o'
    # indicates a new mesh.
    # regex does:
    # '^' : match the first charecter of each newline (with multiline flag)
    # 'o' : match the charecter 'o'
    # '.*': match zero or more of any charecter
    # '\n': up to a newline
    re_obj = '^o.*\n'

    count_vertices = 0
    loaded_data = collections.deque()

    # loop through each sub- mesh
    for text in re.split(re_obj, text, flags=re.MULTILINE):
        if text is None:
            continue

        # find all triangular faces with a regex
        face_tri = ' '.join(re.findall(re_tris,
                                       text)).replace('f',
                                                      ' ').split()
        # convert triangular faces into a numpy array
        face_tri = np.array(face_tri,
                            dtype=np.int64).reshape((-1, 3))

        # find all quad faces with a regex
        face_quad = ' '.join(re.findall(re_quad,
                                        text)).replace('f',
                                                       ' ').split()
        # convert quad faces into a numpy array
        face_quad = np.array(face_quad,
                             dtype=np.int64).reshape((-1, 4))

        # stack the faces into a single (n,3) list
        # triangulate any quad faces
        # this thin wrapper for vstack will ignore empty lists
        faces = util.vstack_empty((face_tri,
                                   geometry.triangulate_quads(face_quad)))

        # If we didn't load any faces, we snagged extra propertie
        # around an object split -- for now, avoid dealing with them.
        if len(faces) == 0:
            continue

        # wavefront has 1- indexed faces, as opposed to 0- indexed
        # additionally, decrement by number of vertices used in prior objects
        faces = faces.astype(np.int64) - 1 - count_vertices

        # find the data with predictable lengths using numpy
        data = np.array(text.split())
        # find the locations of keys, then find the proceeding values
        # indexes which contain vertex information
        vid = np.nonzero(data == 'v')[0].reshape((-1, 1)) + np.arange(3) + 1
        # indexes which contain vertex normal information
        nid = np.nonzero(data == 'vn')[0].reshape((-1, 1)) + np.arange(3) + 1
        # some varients of the format have face groups
        gid = np.nonzero(data == 'g')[0].reshape((-1, 1)) + 1

        # indexes which contain vertex texture information
        tid = np.nonzero(data == 'vt')[0].reshape((-1, 1)) + np.arange(2) + 1

        loaded = {'vertices': data[vid].astype(float),
                  'vertex_normals': data[nid].astype(float),
                  'faces': faces,
                  'metadata': {}}

        # if face groups have been defined add them to metadata
        if len(gid) > 0:
            # indexes which contain face information
            face_key = np.nonzero(data == 'f')[0]
            groups = np.zeros(len(faces), dtype=int)
            for i, g in enumerate(gid):
                groups[np.nonzero(face_key > g)[0]] = i
            loaded['metadata']['face_groups'] = groups

        if len(tid) > 0:
            loaded['metadata']['vertex_texture'] = data[tid].astype(float)

        count_vertices += len(loaded['vertices'])

        loaded_data.append(loaded)

    return list(loaded_data)


def load_msgpack(blob, file_type=None):
    '''
    Load a dict packed with msgpack into kwargs for Trimesh constructor

    Parameters
    ----------
    blob: msgpack packed dict with keys for 'vertices' and 'faces'
    file_type: not used

    Returns
    ----------
    loaded: kwargs for Trimesh constructor (aka mesh=trimesh.Trimesh(**loaded))
    '''

    import msgpack
    if hasattr(blob, 'read'):
        data = msgpack.load(blob)
    else:
        data = msgpack.loads(blob)
    loaded = load_dict(data)
    return loaded


def load_dict(data, file_type=None):
    '''
    Load multiple input types into kwargs for a Trimesh constructor.
    Tries to extract keys ['faces', 'vertices', 'face_normals', 'vertex_normals'].

    Parameters
    ----------
    data: accepts multiple forms
          -dict: has keys for vertices and faces as (n,3) numpy arrays
          -dict: has keys for vertices/faces (n,3) arrays encoded as dicts/base64
                 with trimesh.util.array_to_encoded/trimesh.util.encoded_to_array
          -str:  json blob as dict with either straight array or base64 values
          -file object: json blob of dict
    file_type: not used

    Returns
    -----------
    loaded: dict with keys
            -vertices: (n,3) float
            -faces:    (n,3) int
            -face_normals: (n,3) float (optional)
    '''
    if data is None:
        raise ValueError('data passed to load_dict was None!')
    if util.is_instance_named(data, 'Trimesh'):
        return data
    if util.is_string(data):
        if not '{' in data:
            raise ValueError('Object is not a JSON encoded dictionary!')
        data = json.loads(data.decode('utf-8'))
    elif util.is_file(data):
        data = json.load(data)

    # what shape should the data be to be usable
    mesh_data = {'vertices': (-1, 3),
                 'faces': (-1, (3, 4)),
                 'face_normals': (-1, 3),
                 'face_colors': (-1, (3, 4)),
                 'vertex_normals': (-1, 3),
                 'vertex_colors': (-1, (3, 4))}

    # now go through data structure and if anything is encoded as base64
    # pull it back into numpy arrays
    if util.is_dict(data):
        loaded = {}
        data = util.decode_keys(data, 'utf-8')
        for key, shape in mesh_data.items():
            if key in data:
                loaded[key] = util.encoded_to_array(data[key])
                if not util.is_shape(loaded[key], shape):
                    raise ValueError('Shape of %s is %s, not %s!',
                                     key,
                                     str(loaded[key].shape),
                                     str(shape))
        if len(key) == 0:
            raise ValueError('Unable to extract any mesh data!')
        return loaded
    else:
        raise ValueError('%s object passed to dict loader!',
                         data.__class__.__name__)


_misc_loaders = {'obj': load_wavefront,
                 'off': load_off,
                 'dict': load_dict,
                 'dict64': load_dict,
                 'json': load_dict,
                 'msgpack': load_msgpack}



import numpy as np

from .transformations import rotation_matrix
from .constants import tol, log

from . import util

try:
    from scipy.sparse import coo_matrix
except ImportError:
    log.warning('scipy.sparse.coo_matrix unavailable')


def plane_transform(origin, normal):
    '''
    Given the origin and normal of a plane, find the transform that will move
    that plane to be coplanar with the XY plane

    Parameters
    ----------
    origin: (3,) float, point in space
    normal: (3,) float, plane normal vector

    Returns
    ---------
    transform: (4,4) float, transformation matrix
    '''
    transform = align_vectors(normal, [0, 0, 1])
    transform[0:3, 3] = -np.dot(transform, np.append(origin, 1))[0:3]
    return transform


def transform_around(matrix, point):
    '''
    Given a transformation matrix, apply its rotation component around a
    point in space.

    Parameters
    ----------
    matrix: (4,4) float, transformation matrix
    point:  (3,)  float, point in space

    Returns
    ---------
    result: (4,4) transformation matrix
    '''
    point = np.array(point)
    translate = np.eye(4)
    translate[0:3, 3] = -point
    result = np.dot(matrix, translate)
    translate[0:3, 3] = point
    result = np.dot(translate, result)
    return result


def align_vectors(vector_start, vector_end, return_angle=False):
    '''
    Returns the 4x4 transformation matrix which will rotate from
    vector_start to vector_end, eg:

    vector_end == np.dot(T, np.append(vector_start, 1))[0:3]


    Parameters
    -----------
    vector_start: (3,) float, vector in space
    vector_end:   (3,) float, vector in space
    return_angle: bool, return angle between vectors or not

    Returns
    -----------
    transform: (4,4) float, transformation matrix
    angle:     float, angle in radians (only returned if flag set)

    '''
    start = np.asanyarray(vector_start, dtype=np.float64)
    start /= np.linalg.norm(start)
    end = np.asanyarray(vector_end, dtype=np.float64)
    end /= np.linalg.norm(end)

    cross = np.cross(start, end)
    # we clip the norm to 1, as otherwise floating point bs
    # can cause the arcsin to error
    norm = np.linalg.norm(cross)
    norm = np.clip(norm, -1.0, 1.0)
    direction = np.sign(np.dot(start, end))

    if norm < tol.zero:
        # if the norm is zero, the vectors are the same
        # and no rotation is needed
        T = np.eye(4)
        T[0:3] *= direction
    else:
        angle = np.arcsin(norm)
        if direction < 0:
            angle = np.pi - angle
        T = rotation_matrix(angle, cross)

    check = np.abs(np.dot(T[:3, :3], start) - end)
    if not (check < 1e-5).all():
        raise ValueError('aligning vectors failed!')

    if return_angle:
        return T, angle
    return T


def faces_to_edges(faces, return_index=False):
    '''
    Given a list of faces (n,3), return a list of edges (n*3,2)

    Parameters
    -----------
    faces: (n,3) int, vertex indices representing faces

    Returns
    -----------
    edges: (n*3, 2) int, vertex indices representing edges
    '''
    faces = np.asanyarray(faces)
    edges = np.column_stack((faces[:, (0, 1)],
                             faces[:, (1, 2)],
                             faces[:, (2, 0)])).reshape(-1, 2)
    if return_index:
        face_index = np.tile(np.arange(len(faces)), (3, 1)).T.reshape(-1)
        return edges, face_index
    return edges


def vector_angle(pairs):
    '''
    Find the angles between vector pairs

    Parameters
    ----------
    pairs: (n,2,3) set of vector pairs

    Returns
    ----------
    angles: (n,) float, angles between vectors

    Examples
    ----------
    angles = mesh.face_normals[mesh.face_adjacency]
    '''
    pairs = np.asanyarray(pairs)
    if not util.is_shape(pairs, (-1, 2, 3)):
        raise ValueError('pairs must be (n,2,3)!')
    dots = util.diagonal_dot(pairs[:, 0], pairs[:, 1])
    # clip for floating point error
    dots = np.clip(dots, -1.0, 1.0)
    angles = np.abs(np.arccos(dots))
    return angles


def triangulate_quads(quads):
    '''
    Given a set of quad faces, return them as triangle faces.

    Parameters
    -----------
    quads: (n,4) int, vertex indices of quad faces

    Returns
    -----------
    faces: (m,3) int, vertex indices of triangular faces
    '''
    if len(quads) == 0:
        return quads
    quads = np.asanyarray(quads)
    faces = np.vstack((quads[:, [0, 1, 2]],
                       quads[:, [2, 3, 0]]))
    return faces


def mean_vertex_normals(vertex_count, faces, face_normals, **kwargs):
    '''
    Find vertex normals from the mean of the faces that contain that vertex.

    Parameters
    -----------
    vertex_count: int, the number of vertices faces refer to
    faces:        (n,3) int, list of vertex indices
    face_normals: (n,3) float, normal vector for each face

    Returns
    -----------
    vertex_normals: (vertex_count, 3) float normals for every vertex
                    Uncontained vertices will be zero.
    '''
    def summed_sparse():
        # use a sparse matrix of which face contains each vertex to
        # figure out the summed normal at each vertex
        # allow cached sparse matrix to be passed
        if 'sparse' in kwargs:
            sparse = kwargs['sparse']
        else:
            sparse = index_sparse(vertex_count, faces)
        summed = sparse.dot(face_normals)
        log.debug('Generated vertex normals using sparse matrix')
        return summed

    def summed_loop():
        # loop through every face, in tests was ~50x slower than
        # doing this with a sparse matrix
        summed = np.zeros((vertex_count, 3))
        for face, normal in zip(faces, face_normals):
            summed[face] += normal
        return summed

    try:
        summed = summed_sparse()
    except BaseException:
        log.warning('Unable to generate sparse matrix! Falling back!',
                    exc_info=True)
        summed = summed_loop()
    unit_normals, valid = util.unitize(summed, check_valid=True)
    vertex_normals = np.zeros((vertex_count, 3), dtype=np.float64)
    vertex_normals[valid] = unit_normals

    return vertex_normals


def index_sparse(column_count, indices):
    '''
    Return a sparse matrix for which vertices are contained in which faces.

    Returns
    ---------
    sparse: scipy.sparse.coo_matrix of shape (column_count, len(faces))
            dtype is boolean

    Examples
     ----------
    In [1]: sparse = faces_sparse(len(mesh.vertices), mesh.faces)

    In [2]: sparse.shape
    Out[2]: (12, 20)

    In [3]: mesh.faces.shape
    Out[3]: (20, 3)

    In [4]: mesh.vertices.shape
    Out[4]: (12, 3)

    In [5]: dense = sparse.toarray().astype(int)

    In [6]: dense
    Out[6]:
    array([[1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
           [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0],
           [0, 0, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 1, 0, 0],
           [0, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
           [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 1],
           [1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0],
           [0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0],
           [0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1],
           [0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1]])

    In [7]: dense.sum(axis=0)
    Out[7]: array([3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3])
    '''
    indices = np.asanyarray(indices)
    column_count = int(column_count)

    row = indices.reshape(-1)
    col = np.tile(np.arange(len(indices)).reshape(
        (-1, 1)), (1, indices.shape[1])).reshape(-1)

    shape = (column_count, len(indices))
    data = np.ones(len(col), dtype=np.bool)
    sparse = coo_matrix((data, (row, col)),
                        shape=shape,
                        dtype=np.bool)
    return sparse


def medial_axis(samples, contains):
    '''
    Given a set of samples on a boundary, find the approximate medial axis based
    on a voronoi diagram and a containment function which can assess whether
    a point is inside or outside of the closed geometry.

    Parameters
    ----------
    samples:    (n,d) set of points on the boundary of the geometry
    contains:   function which takes (m,d) points and returns an (m) bool array

    Returns
    ----------
    lines:     (n,2,2) set of line segments
    '''

    from scipy.spatial import Voronoi
    from .path.io.load import load_path

    # create the voronoi diagram, after vertically stacking the points
    # deque from a sequnce into a clean (m,2) array
    voronoi = Voronoi(samples)
    # which voronoi vertices are contained inside the original polygon
    contained = contains(voronoi.vertices)
    # ridge vertices of -1 are outside, make sure they are False
    contained = np.append(contained, False)
    inside = [i for i in voronoi.ridge_vertices if contained[i].all()]
    line_indices = np.vstack([util.stack_lines(i)
                              for i in inside if len(i) >= 2])
    lines = voronoi.vertices[line_indices]
    return load_path(lines)
